# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from email import message_from_string
from email.utils import mktime_tz
from email.utils import parsedate_tz

from twisted.trial import unittest

from buildbot.changes.mail import CVSMaildirSource

#
# Sample message from CVS version 1.11
#

cvs1_11_msg = """From: Andy Howell <andy@example.com>
To: buildbot@example.com
Subject: cvs module MyModuleName
Date: Sat, 07 Aug 2010 11:11:49 +0000
X-Mailer: Python buildbot-cvs-mail $Revision: 1.3 $

Cvsmode: 1.11
Category: None
CVSROOT: :ext:cvshost.example.com:/cvsroot
Files: base/module/src/make GNUmakefile,1.362,1.363
Project: MyModuleName
Update of /cvsroot/base/module/src/make
In directory cvshost:/tmp/cvs-serv10922

Modified Files:
        GNUmakefile
Log Message:
Commented out some stuff.
"""

#
# Sample message from CVS version 1.12
#
# Paths are handled differently by the two versions
#

cvs1_12_msg = """Date: Wed, 11 Aug 2010 04:56:44 +0000
From: andy@example.com
To: buildbot@example.com
Subject: cvs update for project RaiCore
X-Mailer: Python buildbot-cvs-mail $Revision: 1.3 $

Cvsmode: 1.12
Category: None
CVSROOT: :ext:cvshost.example.com:/cvsroot
Files: file1.cpp 1.77 1.78 file2.cpp 1.75 1.76
Path: base/module/src
Project: MyModuleName
Update of /cvsroot/base/module/src
In directory example.com:/tmp/cvs-serv26648/InsightMonAgent

Modified Files:
        file1.cpp file2.cpp
Log Message:
Changes for changes sake
"""


class TestCVSMaildirSource(unittest.TestCase):
    def test_CVSMaildirSource_create_change_from_cvs1_11msg(self):
        m = message_from_string(cvs1_11_msg)
        src = CVSMaildirSource('/dev/null')
        src, chdict = src.parse(m)
        self.assertNotEqual(chdict, None)
        self.assertEqual(chdict['author'], 'andy')
        self.assertEqual(len(chdict['files']), 1)
        self.assertEqual(chdict['files'][0], 'base/module/src/make/GNUmakefile')
        self.assertEqual(chdict['comments'], 'Commented out some stuff.\n')
        self.assertFalse(chdict['isdir'])
        self.assertEqual(chdict['revision'], '2010-08-07 11:11:49')
        dateTuple = parsedate_tz('Sat, 07 Aug 2010 11:11:49 +0000')
        self.assertEqual(chdict['when'], mktime_tz(dateTuple))
        self.assertEqual(chdict['branch'], None)
        self.assertEqual(chdict['repository'], ':ext:cvshost.example.com:/cvsroot')
        self.assertEqual(chdict['project'], 'MyModuleName')
        self.assertEqual(len(chdict['properties']), 0)
        self.assertEqual(src, 'cvs')

    def test_CVSMaildirSource_create_change_from_cvs1_12msg(self):
        m = message_from_string(cvs1_12_msg)
        src = CVSMaildirSource('/dev/null')
        src, chdict = src.parse(m)
        self.assertNotEqual(chdict, None)
        self.assertEqual(chdict['author'], 'andy')
        self.assertEqual(len(chdict['files']), 2)
        self.assertEqual(chdict['files'][0], 'base/module/src/file1.cpp')
        self.assertEqual(chdict['files'][1], 'base/module/src/file2.cpp')
        self.assertEqual(chdict['comments'], 'Changes for changes sake\n')
        self.assertFalse(chdict['isdir'])
        self.assertEqual(chdict['revision'], '2010-08-11 04:56:44')
        dateTuple = parsedate_tz('Wed, 11 Aug 2010 04:56:44 +0000')
        self.assertEqual(chdict['when'], mktime_tz(dateTuple))
        self.assertEqual(chdict['branch'], None)
        self.assertEqual(chdict['repository'], ':ext:cvshost.example.com:/cvsroot')
        self.assertEqual(chdict['project'], 'MyModuleName')
        self.assertEqual(len(chdict['properties']), 0)
        self.assertEqual(src, 'cvs')

    def test_CVSMaildirSource_create_change_from_cvs1_12_with_no_path(self):
        msg = cvs1_12_msg.replace('Path: base/module/src', '')
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        try:
            assert src.parse(m)[1]
        except ValueError:
            pass
        else:
            self.fail('Expect ValueError.')

    def test_CVSMaildirSource_create_change_with_bad_cvsmode(self):
        # Branch is indicated after 'Tag:' in modified file list
        msg = cvs1_11_msg.replace('Cvsmode: 1.11', 'Cvsmode: 9.99')
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        try:
            assert src.parse(m)[1]
        except ValueError:
            pass
        else:
            self.fail('Expected ValueError')

    def test_CVSMaildirSource_create_change_with_branch(self):
        # Branch is indicated after 'Tag:' in modified file list
        msg = cvs1_11_msg.replace(
            '        GNUmakefile', '      Tag: Test_Branch\n      GNUmakefile'
        )
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        chdict = src.parse(m)[1]
        self.assertEqual(chdict['branch'], 'Test_Branch')

    def test_CVSMaildirSource_create_change_with_category(self):
        msg = cvs1_11_msg.replace('Category: None', 'Category: Test category')
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        chdict = src.parse(m)[1]
        self.assertEqual(chdict['category'], 'Test category')

    def test_CVSMaildirSource_create_change_with_no_comment(self):
        # Strip off comments
        msg = cvs1_11_msg[: cvs1_11_msg.find('Commented out some stuff')]
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        chdict = src.parse(m)[1]
        self.assertEqual(chdict['comments'], None)

    def test_CVSMaildirSource_create_change_with_no_files(self):
        # A message with no files is likely not for us
        msg = cvs1_11_msg.replace('Files: base/module/src/make GNUmakefile,1.362,1.363', '')
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        chdict = src.parse(m)
        self.assertEqual(chdict, None)

    def test_CVSMaildirSource_create_change_with_no_project(self):
        msg = cvs1_11_msg.replace('Project: MyModuleName', '')
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        chdict = src.parse(m)[1]
        self.assertEqual(chdict['project'], None)

    def test_CVSMaildirSource_create_change_with_no_repository(self):
        msg = cvs1_11_msg.replace('CVSROOT: :ext:cvshost.example.com:/cvsroot', '')
        m = message_from_string(msg)
        src = CVSMaildirSource('/dev/null')
        chdict = src.parse(m)[1]
        self.assertEqual(chdict['repository'], None)

    def test_CVSMaildirSource_create_change_with_property(self):
        m = message_from_string(cvs1_11_msg)
        propDict = {'foo': 'bar'}
        src = CVSMaildirSource('/dev/null', properties=propDict)
        chdict = src.parse(m)[1]
        self.assertEqual(chdict['properties']['foo'], 'bar')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
password store based provider
"""

from __future__ import annotations

import os
from pathlib import Path

from twisted.internet import defer

from buildbot import config
from buildbot.secrets.providers.base import SecretProviderBase
from buildbot.util import runprocess


class SecretInPass(SecretProviderBase):
    """
    secret is stored in a password store
    """

    name: str | None = "SecretInPass"  # type: ignore[assignment]

    def checkPassIsInPath(self):
        if not any((Path(p) / "pass").is_file() for p in os.environ["PATH"].split(":")):
            config.error("pass does not exist in PATH")

    def checkPassDirectoryIsAvailableAndReadable(self, dirname):
        if not os.access(dirname, os.F_OK):
            config.error(f"directory {dirname} does not exist")

    def checkConfig(self, gpgPassphrase=None, dirname=None):
        self.checkPassIsInPath()
        if dirname:
            self.checkPassDirectoryIsAvailableAndReadable(dirname)

    def reconfigService(self, gpgPassphrase=None, dirname=None):
        self._env = {**os.environ}
        if gpgPassphrase:
            self._env["PASSWORD_STORE_GPG_OPTS"] = f"--passphrase {gpgPassphrase}"
        if dirname:
            self._env["PASSWORD_STORE_DIR"] = dirname

    @defer.inlineCallbacks
    def get(self, entry):
        """
        get the value from pass identified by 'entry'
        """
        try:
            rc, output = yield runprocess.run_process(
                self.master.reactor,
                ['pass', entry],
                env=self._env,
                collect_stderr=False,
                stderr_is_error=True,
            )
            if rc != 0:
                return None
            return output.decode("utf-8", "ignore").splitlines()[0]
        except OSError:
            return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot import config
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps import maxq
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin


class TestShellCommandExecution(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_testdir_required(self):
        with self.assertRaises(config.ConfigErrors):
            maxq.MaxQ()

    def test_success(self):
        self.setup_step(maxq.MaxQ(testdir='x'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=["run_maxq.py", "x"])
            .stdout('no failures\n')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='success')
        return self.run_step()

    def test_nonzero_rc_no_failures(self):
        self.setup_step(maxq.MaxQ(testdir='x'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=["run_maxq.py", "x"])
            .stdout('no failures\n')
            .exit(2)
        )
        self.expect_outcome(result=FAILURE, state_string='1 maxq failures')
        return self.run_step()

    def test_failures(self):
        self.setup_step(maxq.MaxQ(testdir='x'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=["run_maxq.py", "x"])
            .stdout('\nTEST FAILURE: foo\n' * 10)
            .exit(2)
        )
        self.expect_outcome(result=FAILURE, state_string='10 maxq failures')
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from collections import OrderedDict
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.process.properties import Interpolate
from buildbot.process.results import SUCCESS
from buildbot.steps.package.rpm import rpmbuild
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin


class RpmBuild(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_no_specfile(self):
        with self.assertRaises(config.ConfigErrors):
            rpmbuild.RpmBuild()

    def test_success(self):
        self.setup_step(rpmbuild.RpmBuild(specfile="foo.spec", dist=".el5"))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='rpmbuild --define "_topdir '
                '`pwd`" --define "_builddir `pwd`" --define "_rpmdir '
                '`pwd`" --define "_sourcedir `pwd`" --define "_specdir '
                '`pwd`" --define "_srcrpmdir `pwd`" --define "dist .el5" '
                '-ba foo.spec',
            )
            .stdout('lalala')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='RPMBUILD')
        return self.run_step()

    @mock.patch('builtins.open', mock.mock_open())
    def test_autoRelease(self):
        self.setup_step(rpmbuild.RpmBuild(specfile="foo.spec", autoRelease=True))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='rpmbuild --define "_topdir '
                '`pwd`" --define "_builddir `pwd`" --define "_rpmdir `pwd`" '
                '--define "_sourcedir `pwd`" --define "_specdir `pwd`" '
                '--define "_srcrpmdir `pwd`" --define "_release 0" '
                '--define "dist .el6" -ba foo.spec',
            )
            .stdout('Your code has been rated at 10/10')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='RPMBUILD')
        return self.run_step()

    def test_define(self):
        defines = [("a", "1"), ("b", "2")]
        self.setup_step(rpmbuild.RpmBuild(specfile="foo.spec", define=OrderedDict(defines)))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='rpmbuild --define "_topdir '
                '`pwd`" --define "_builddir `pwd`" --define "_rpmdir '
                '`pwd`" --define "_sourcedir `pwd`" --define '
                '"_specdir `pwd`" --define "_srcrpmdir `pwd`" '
                '--define "a 1" --define "b 2" --define "dist .el6" '
                '-ba foo.spec',
            )
            .stdout('Your code has been rated at 10/10')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='RPMBUILD')
        return self.run_step()

    def test_define_none(self):
        self.setup_step(rpmbuild.RpmBuild(specfile="foo.spec", define=None))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='rpmbuild --define "_topdir '
                '`pwd`" --define "_builddir `pwd`" --define "_rpmdir '
                '`pwd`" --define "_sourcedir `pwd`" --define '
                '"_specdir `pwd`" --define "_srcrpmdir `pwd`" '
                '--define "dist .el6" -ba foo.spec',
            )
            .stdout('Your code has been rated at 10/10')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='RPMBUILD')
        return self.run_step()

    @defer.inlineCallbacks
    def test_renderable_dist(self):
        self.setup_step(
            rpmbuild.RpmBuild(specfile="foo.spec", dist=Interpolate('%(prop:renderable_dist)s'))
        )
        self.build.setProperty('renderable_dist', '.el7', 'test')
        self.expect_commands(
            ExpectShell(
                workdir='wkdir',
                command='rpmbuild --define "_topdir '
                '`pwd`" --define "_builddir `pwd`" --define "_rpmdir '
                '`pwd`" --define "_sourcedir `pwd`" --define "_specdir '
                '`pwd`" --define "_srcrpmdir `pwd`" --define "dist .el7" '
                '-ba foo.spec',
            )
            .stdout('lalala')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string='RPMBUILD')
        yield self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from unittest import mock

from twisted.internet import defer
from twisted.python import failure
from twisted.trial import unittest

from buildbot.data import changesources
from buildbot.db.changesources import ChangeSourceAlreadyClaimedError
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces


class ChangeSourceEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = changesources.ChangeSourceEndpoint
    resourceTypeClass = changesources.ChangeSource

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()

        yield self.master.db.insert_test_data([
            fakedb.Master(id=22, active=0),
            fakedb.Master(id=33, active=1),
            fakedb.ChangeSource(id=13, name='some:changesource'),
            fakedb.ChangeSource(id=14, name='other:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=14, masterid=22),
            fakedb.ChangeSource(id=15, name='another:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=15, masterid=33),
        ])

    @defer.inlineCallbacks
    def test_get_existing(self):
        """get an existing changesource by id"""
        changesource = yield self.callGet(('changesources', 14))

        self.validateData(changesource)
        self.assertEqual(changesource['name'], 'other:changesource')

    @defer.inlineCallbacks
    def test_get_no_master(self):
        """get a changesource with no master"""
        changesource = yield self.callGet(('changesources', 13))

        self.validateData(changesource)
        self.assertEqual(changesource['master'], None)

    @defer.inlineCallbacks
    def test_get_masterid_existing(self):
        """get an existing changesource by id on certain master"""
        changesource = yield self.callGet(('masters', 22, 'changesources', 14))

        self.validateData(changesource)
        self.assertEqual(changesource['name'], 'other:changesource')

    @defer.inlineCallbacks
    def test_get_masterid_no_match(self):
        """get an existing changesource by id on the wrong master"""
        changesource = yield self.callGet(('masters', 33, 'changesources', 13))

        self.assertEqual(changesource, None)

    @defer.inlineCallbacks
    def test_get_masterid_missing(self):
        """get an existing changesource by id on an invalid master"""
        changesource = yield self.callGet(('masters', 25, 'changesources', 13))

        self.assertEqual(changesource, None)

    @defer.inlineCallbacks
    def test_get_missing(self):
        """get an invalid changesource by id"""
        changesource = yield self.callGet(('changesources', 99))

        self.assertEqual(changesource, None)


class ChangeSourcesEndpoint(endpoint.EndpointMixin, unittest.TestCase):
    endpointClass = changesources.ChangeSourcesEndpoint
    resourceTypeClass = changesources.ChangeSource

    @defer.inlineCallbacks
    def setUp(self):
        yield self.setUpEndpoint()
        yield self.master.db.insert_test_data([
            fakedb.Master(id=22, active=0),
            fakedb.Master(id=33, active=1),
            fakedb.ChangeSource(id=13, name='some:changesource'),
            fakedb.ChangeSource(id=14, name='other:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=14, masterid=22),
            fakedb.ChangeSource(id=15, name='another:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=15, masterid=33),
            fakedb.ChangeSource(id=16, name='wholenother:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=16, masterid=33),
        ])

    @defer.inlineCallbacks
    def test_get(self):
        changesources = yield self.callGet(('changesources',))

        for cs in changesources:
            self.validateData(cs)

        self.assertEqual(sorted([m['changesourceid'] for m in changesources]), [13, 14, 15, 16])

    @defer.inlineCallbacks
    def test_get_masterid(self):
        changesources = yield self.callGet(('masters', 33, 'changesources'))

        for cs in changesources:
            self.validateData(cs)

        self.assertEqual(sorted([m['changesourceid'] for m in changesources]), [15, 16])

    @defer.inlineCallbacks
    def test_get_masterid_missing(self):
        changesources = yield self.callGet(('masters', 23, 'changesources'))

        self.assertEqual(changesources, [])


class ChangeSource(TestReactorMixin, interfaces.InterfaceTests, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantDb=True, wantData=True)
        self.rtype = changesources.ChangeSource(self.master)

    def test_signature_findChangeSourceId(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.findChangeSourceId,  # fake
            self.rtype.findChangeSourceId,
        )  # real
        def findChangeSourceId(self, name):
            pass

    @defer.inlineCallbacks
    def test_findChangeSourceId(self):
        self.master.db.changesources.findChangeSourceId = mock.Mock(return_value=defer.succeed(10))
        self.assertEqual((yield self.rtype.findChangeSourceId('cs')), 10)
        self.master.db.changesources.findChangeSourceId.assert_called_with('cs')

    def test_signature_trySetChangeSourceMaster(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.trySetChangeSourceMaster,  # fake
            self.rtype.trySetChangeSourceMaster,
        )  # real
        def trySetChangeSourceMaster(self, changesourceid, masterid):
            pass

    @defer.inlineCallbacks
    def test_trySetChangeSourceMaster_succeeds(self):
        self.master.db.changesources.setChangeSourceMaster = mock.Mock(
            return_value=defer.succeed(None)
        )
        yield self.rtype.trySetChangeSourceMaster(10, 20)
        self.master.db.changesources.setChangeSourceMaster.assert_called_with(10, 20)

    @defer.inlineCallbacks
    def test_trySetChangeSourceMaster_fails(self):
        d = defer.fail(failure.Failure(ChangeSourceAlreadyClaimedError('oh noes')))

        self.master.db.changesources.setChangeSourceMaster = mock.Mock(return_value=d)
        result = yield self.rtype.trySetChangeSourceMaster(10, 20)

        self.assertFalse(result)

    @defer.inlineCallbacks
    def test_trySetChangeSourceMaster_raisesOddException(self):
        d = defer.fail(failure.Failure(RuntimeError('oh noes')))

        self.master.db.changesources.setChangeSourceMaster = mock.Mock(return_value=d)

        try:
            yield self.rtype.trySetChangeSourceMaster(10, 20)
        except RuntimeError:
            pass
        else:
            self.fail("The RuntimeError did not propagate")

    @defer.inlineCallbacks
    def test__masterDeactivated(self):
        yield self.master.db.insert_test_data([
            fakedb.Master(id=22, active=0),
            fakedb.ChangeSource(id=13, name='some:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=13, masterid=22),
            fakedb.ChangeSource(id=14, name='other:changesource'),
            fakedb.ChangeSourceMaster(changesourceid=14, masterid=22),
        ])
        yield self.rtype._masterDeactivated(22)
        self.assertIsNone((yield self.master.db.changesources.get_change_source_master(13)))
        self.assertIsNone((yield self.master.db.changesources.get_change_source_master(14)))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.process.results import SUCCESS
from buildbot.steps.source import gerrit
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectListdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import config
from buildbot.test.util import sourcesteps


class TestGerrit(
    sourcesteps.SourceStepMixin, config.ConfigErrorsMixin, TestReactorMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_mode_full_clean(self):
        self.setup_step(
            gerrit.Gerrit(
                repourl='http://github.com/buildbot/buildbot.git', mode='full', method='clean'
            )
        )
        self.build.setProperty("event.change.project", "buildbot")
        self.sourcestamp.project = 'buildbot'
        self.build.setProperty("event.patchSet.ref", "gerrit_branch")

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', '--version'])
            .stdout('git version 1.7.5')
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectListdir(dir='wkdir').files(['.git']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'clean', '-f', '-f', '-d']).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'git',
                    'fetch',
                    '-f',
                    '--progress',
                    'http://github.com/buildbot/buildbot.git',
                    'gerrit_branch',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-f', 'FETCH_HEAD']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-B', 'gerrit_branch']).exit(
                0
            ),
            ExpectShell(workdir='wkdir', command=['git', 'rev-parse', 'HEAD'])
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'f6ad368298bd941e934a41f3babc827b2aa95a1d', 'Gerrit')
        return self.run_step()

    def test_mode_full_clean_force_build(self):
        self.setup_step(
            gerrit.Gerrit(
                repourl='http://github.com/buildbot/buildbot.git', mode='full', method='clean'
            )
        )
        self.build.setProperty("event.change.project", "buildbot")
        self.sourcestamp.project = 'buildbot'
        self.build.setProperty("gerrit_change", "1234/567")

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', '--version'])
            .stdout('git version 1.7.5')
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectListdir(dir='wkdir').files(['.git']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'clean', '-f', '-f', '-d']).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'git',
                    'fetch',
                    '-f',
                    '--progress',
                    'http://github.com/buildbot/buildbot.git',
                    'refs/changes/34/1234/567',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-f', 'FETCH_HEAD']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['git', 'checkout', '-B', 'refs/changes/34/1234/567']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'rev-parse', 'HEAD'])
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'f6ad368298bd941e934a41f3babc827b2aa95a1d', 'Gerrit')
        return self.run_step()

    def test_mode_full_clean_force_same_project(self):
        self.setup_step(
            gerrit.Gerrit(
                repourl='http://github.com/buildbot/buildbot.git',
                mode='full',
                method='clean',
                codebase='buildbot',
            )
        )
        self.build.setProperty("event.change.project", "buildbot")
        self.sourcestamp.project = 'buildbot'
        self.build.setProperty("gerrit_change", "1234/567")

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', '--version'])
            .stdout('git version 1.7.5')
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectListdir(dir='wkdir').files(['.git']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'clean', '-f', '-f', '-d']).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'git',
                    'fetch',
                    '-f',
                    '--progress',
                    'http://github.com/buildbot/buildbot.git',
                    'refs/changes/34/1234/567',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-f', 'FETCH_HEAD']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['git', 'checkout', '-B', 'refs/changes/34/1234/567']
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'rev-parse', 'HEAD'])
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property(
            'got_revision', {'buildbot': 'f6ad368298bd941e934a41f3babc827b2aa95a1d'}, 'Gerrit'
        )
        return self.run_step()

    def test_mode_full_clean_different_project(self):
        self.setup_step(
            gerrit.Gerrit(
                repourl='http://github.com/buildbot/buildbot.git',
                mode='full',
                method='clean',
                codebase='buildbot',
            )
        )
        self.build.setProperty("event.change.project", "buildbot")
        self.sourcestamp.project = 'not_buildbot'
        self.build.setProperty("gerrit_change", "1234/567")

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', '--version'])
            .stdout('git version 1.7.5')
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectListdir(dir='wkdir').files(['.git']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'clean', '-f', '-f', '-d']).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'git',
                    'fetch',
                    '-f',
                    '--progress',
                    'http://github.com/buildbot/buildbot.git',
                    'HEAD',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-f', 'FETCH_HEAD']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'rev-parse', 'HEAD'])
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial.unittest import TestCase

from buildbot.test.util.db import get_trial_parallel_from_cwd


class Tests(TestCase):
    def test_unknown(self) -> None:
        self.assertIsNone(get_trial_parallel_from_cwd(""))
        self.assertIsNone(get_trial_parallel_from_cwd("/"))
        self.assertIsNone(get_trial_parallel_from_cwd("/abc"))
        self.assertIsNone(get_trial_parallel_from_cwd("/abc/"))
        self.assertIsNone(get_trial_parallel_from_cwd("/abc/abc/1"))
        self.assertIsNone(get_trial_parallel_from_cwd("/abc/abc/1/"))
        self.assertIsNone(get_trial_parallel_from_cwd("/_trial_temp/abc/1"))
        self.assertIsNone(get_trial_parallel_from_cwd("/_trial_temp/abc/1/"))

    def test_single(self) -> None:
        self.assertIs(get_trial_parallel_from_cwd("_trial_temp"), False)
        self.assertIs(get_trial_parallel_from_cwd("_trial_temp/"), False)
        self.assertIs(get_trial_parallel_from_cwd("/_trial_temp"), False)
        self.assertIs(get_trial_parallel_from_cwd("/_trial_temp/"), False)
        self.assertIs(get_trial_parallel_from_cwd("/abc/_trial_temp"), False)
        self.assertIs(get_trial_parallel_from_cwd("/abc/_trial_temp/"), False)

    def test_index(self) -> None:
        self.assertEqual(get_trial_parallel_from_cwd("_trial_temp/0"), 0)
        self.assertEqual(get_trial_parallel_from_cwd("_trial_temp/0/"), 0)
        self.assertEqual(get_trial_parallel_from_cwd("_trial_temp/5"), 5)
        self.assertEqual(get_trial_parallel_from_cwd("_trial_temp/5/"), 5)
        self.assertEqual(get_trial_parallel_from_cwd("/_trial_temp/0"), 0)
        self.assertEqual(get_trial_parallel_from_cwd("/_trial_temp/0/"), 0)
        self.assertEqual(get_trial_parallel_from_cwd("/_trial_temp/5"), 5)
        self.assertEqual(get_trial_parallel_from_cwd("/_trial_temp/5/"), 5)
        self.assertEqual(get_trial_parallel_from_cwd("abc/_trial_temp/0"), 0)
        self.assertEqual(get_trial_parallel_from_cwd("abc/_trial_temp/0/"), 0)
        self.assertEqual(get_trial_parallel_from_cwd("abc/_trial_temp/5"), 5)
        self.assertEqual(get_trial_parallel_from_cwd("abc/_trial_temp/5/"), 5)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process import properties
from buildbot.schedulers import timed
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler
from buildbot.test.util.state import StateTestMixin


class NightlyTriggerable(
    scheduler.SchedulerMixin, TestReactorMixin, StateTestMixin, unittest.TestCase
):
    SCHEDULERID = 327
    OBJECTID = 1327

    @defer.inlineCallbacks
    def makeScheduler(self, firstBuildDuration=0, **kwargs):
        sched = yield self.attachScheduler(
            timed.NightlyTriggerable(**kwargs),
            self.OBJECTID,
            self.SCHEDULERID,
            overrideBuildsetMethods=True,
            createBuilderDB=True,
        )
        return sched

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    # utilities

    def assertBuildsetAdded(self, sourcestamps=None, properties=None):
        if sourcestamps is None:
            sourcestamps = {}
        if properties is None:
            properties = {}
        properties['scheduler'] = ('test', 'Scheduler')
        self.assertEqual(
            self.addBuildsetCalls,
            [
                (
                    'addBuildsetForSourceStampsWithDefaults',
                    {
                        "builderNames": None,  # uses the default
                        "priority": None,
                        "properties": properties,
                        "reason": "The NightlyTriggerable scheduler named 'test' triggered this build",
                        "sourcestamps": sourcestamps,
                        "waited_for": False,
                    },
                ),
            ],
        )
        self.addBuildsetCalls = []

    def assertNoBuildsetAdded(self):
        self.assertEqual(self.addBuildsetCalls, [])

    # tests

    @defer.inlineCallbacks
    def test_constructor_no_reason(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'])
        yield sched.configureService()
        self.assertEqual(
            sched.reason, "The NightlyTriggerable scheduler named 'test' triggered this build"
        )

    @defer.inlineCallbacks
    def test_constructor_reason(self):
        sched = yield self.makeScheduler(
            name='test', builderNames=['test'], reason="hourlytriggerable"
        )
        yield sched.configureService()
        self.assertEqual(sched.reason, "hourlytriggerable")

    @defer.inlineCallbacks
    def test_constructor_month(self):
        sched = yield self.makeScheduler(name='test', builderNames=['test'], month='1')
        yield sched.configureService()
        self.assertEqual(sched.month, "1")

    @defer.inlineCallbacks
    def test_timer_noBuilds(self):
        yield self.makeScheduler(name='test', builderNames=['test'], minute=[5])
        yield self.master.startService()

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertEqual(self.addBuildsetCalls, [])

    @defer.inlineCallbacks
    def test_timer_oneTrigger(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )
        yield self.master.startService()

        sched.trigger(
            False,
            [
                {
                    "revision": 'myrev',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "codebase": 'cb',
                },
            ],
            set_props=None,
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev',
                },
            ]
        )

    @defer.inlineCallbacks
    def test_timer_twoTriggers(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )

        yield self.master.startService()

        sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev1',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                }
            ],
            set_props=None,
        )
        sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev2',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                }
            ],
            set_props=None,
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    # builds the second trigger's revision
                    "revision": 'myrev2',
                },
            ]
        )

    @defer.inlineCallbacks
    def test_timer_oneTrigger_then_noBuild(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )

        yield self.master.startService()

        sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                }
            ],
            set_props=None,
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev',
                },
            ]
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        # no trigger, so the second did not build
        self.assertNoBuildsetAdded()

    @defer.inlineCallbacks
    def test_timer_oneTriggers_then_oneTrigger(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )

        yield self.master.startService()

        sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev1',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                }
            ],
            set_props=None,
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev1',
                },
            ]
        )

        sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev2',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                }
            ],
            set_props=None,
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev2',
                },
            ]
        )

    @defer.inlineCallbacks
    def test_savedTrigger(self):
        yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )

        value_json = (
            '[ [ {"codebase": "cb", "project": "p", "repository": "r", '
            '"branch": "br", "revision": "myrev"} ], {}, null, null ]'
        )

        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.SCHEDULERID, name='test', class_name='NightlyTriggerable'),
            fakedb.ObjectState(
                objectid=self.SCHEDULERID, name='lastTrigger', value_json=value_json
            ),
        ])

        yield self.master.startService()

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev',
                },
            ]
        )

    @defer.inlineCallbacks
    def test_savedTrigger_dict(self):
        yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )

        value_json = (
            '[ { "cb": {"codebase": "cb", "project": "p", "repository": "r", '
            '"branch": "br", "revision": "myrev"} }, {}, null, null ]'
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.SCHEDULERID, name='test', class_name='NightlyTriggerable'),
            fakedb.ObjectState(
                objectid=self.SCHEDULERID, name='lastTrigger', value_json=value_json
            ),
        ])

        yield self.master.startService()

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev',
                },
            ]
        )

    @defer.inlineCallbacks
    def test_saveTrigger(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.SCHEDULERID, name='test', class_name='NightlyTriggerable'),
        ])

        yield self.master.startService()

        _, d = sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                },
            ],
            set_props=None,
        )

        yield d
        yield self.assert_state(
            self.SCHEDULERID,
            lastTrigger=[
                [
                    {
                        "codebase": 'cb',
                        "revision": 'myrev',
                        "branch": 'br',
                        "project": 'p',
                        "repository": 'r',
                    },
                ],
                {},
                None,
                None,
            ],
        )

    @defer.inlineCallbacks
    def test_saveTrigger_noTrigger(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.SCHEDULERID, name='test', class_name='NightlyTriggerable'),
        ])

        yield self.master.startService()

        _, d = sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                },
            ],
            set_props=None,
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        yield d
        yield self.assert_state(self.SCHEDULERID, lastTrigger=None)

    @defer.inlineCallbacks
    def test_triggerProperties(self):
        sched = yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.SCHEDULERID, name='test', class_name='NightlyTriggerable'),
        ])

        yield self.master.startService()

        sched.trigger(
            False,
            [
                {
                    "codebase": 'cb',
                    "revision": 'myrev',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                },
            ],
            properties.Properties(testprop='test'),
        )

        yield self.assert_state(
            self.SCHEDULERID,
            lastTrigger=[
                [
                    {
                        "codebase": 'cb',
                        "revision": 'myrev',
                        "branch": 'br',
                        "project": 'p',
                        "repository": 'r',
                    },
                ],
                {'testprop': ['test', 'TEST']},
                None,
                None,
            ],
        )

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            properties={"testprop": ('test', 'TEST')},
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev',
                },
            ],
        )

    @defer.inlineCallbacks
    def test_savedProperties(self):
        yield self.makeScheduler(
            name='test',
            builderNames=['test'],
            minute=[5],
            codebases={'cb': {'repository': 'annoying'}},
        )

        value_json = (
            '[ [ {"codebase": "cb", "project": "p", "repository": "r", '
            '"branch": "br", "revision": "myrev"} ], '
            '{"testprop": ["test", "TEST"]}, null, null ]'
        )
        yield self.master.db.insert_test_data([
            fakedb.Object(id=self.SCHEDULERID, name='test', class_name='NightlyTriggerable'),
            fakedb.ObjectState(
                objectid=self.SCHEDULERID, name='lastTrigger', value_json=value_json
            ),
        ])

        yield self.master.startService()

        self.reactor.advance(60 * 60)  # Run for 1h

        self.assertBuildsetAdded(
            properties={'testprop': ('test', 'TEST')},
            sourcestamps=[
                {
                    "codebase": 'cb',
                    "branch": 'br',
                    "project": 'p',
                    "repository": 'r',
                    "revision": 'myrev',
                },
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This module is left for backward compatibility of old-named worker API.
# It should never be imported by Buildbot.
from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # NOTE: can't use newer `|` annotations for TypedDict def
    from typing import Optional
    from typing import Union

    from typing_extensions import Literal
    from typing_extensions import TypedDict

    # Use alternative synthax due to some fields containing non-valid indentifier chars (eg. '-')
    Options = TypedDict(
        "Options",
        {
            # flags
            "no-logrotate": bool,
            "relocatable": bool,
            "quiet": bool,
            "use-tls": bool,
            "delete-leftover-dirs": bool,
            # options
            "basedir": str,
            "allow-shutdown": Optional[Literal["signal"]],
            "umask": Union[int, str, None],
            "log-size": int,
            "log-count": Union[int, str],
            "keepalive": int,
            "maxdelay": int,
            "numcpus": Union[int, str, None],
            "protocol": Literal["pb", "msgpack", "null"],
            "maxretries": Union[int, str, None],
            "connection-string": Union[str, None],
            "proxy-connection-string": Union[str, None],
            # arguments
            "host": Union[str, None],
            "port": Union[int, None],
            "name": str,
            "passwd": str,
        },
    )


def isWorkerDir(dir: str) -> bool:
    def print_error(error_message: str) -> None:
        print(f"{error_message}\ninvalid worker directory '{dir}'")

    buildbot_tac = os.path.join(dir, "buildbot.tac")
    try:
        with open(buildbot_tac) as f:
            contents = f.read()
    except OSError as exception:
        print_error(f"error reading '{buildbot_tac}': {exception.strerror}")
        return False

    if "Application('buildbot-worker')" not in contents:
        print_error(f"unexpected content in '{buildbot_tac}'")
        return False

    return True



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


"""Base classes handy for use with PB clients."""

from twisted.internet import protocol
from twisted.python import log
from twisted.spread import pb
from twisted.spread.pb import PBClientFactory

from buildbot.util import bytes2unicode


class NewCredPerspective(pb.Avatar):
    def attached(self, mind):
        return self

    def detached(self, mind):
        pass


class ReconnectingPBClientFactory(PBClientFactory, protocol.ReconnectingClientFactory):
    """Reconnecting client factory for PB brokers.

    Like PBClientFactory, but if the connection fails or is lost, the factory
    will attempt to reconnect.

    Instead of using f.getRootObject (which gives a Deferred that can only
    be fired once), override the gotRootObject method.

    Instead of using the newcred f.login (which is also one-shot), call
    f.startLogin() with the credentials and client, and override the
    gotPerspective method.

    Instead of using the oldcred f.getPerspective (also one-shot), call
    f.startGettingPerspective() with the same arguments, and override
    gotPerspective.

    gotRootObject and gotPerspective will be called each time the object is
    received (once per successful connection attempt). You will probably want
    to use obj.notifyOnDisconnect to find out when the connection is lost.

    If an authorization error occurs, failedToGetPerspective() will be
    invoked.

    To use me, subclass, then hand an instance to a connector (like
    TCPClient).
    """

    def __init__(self):
        super().__init__()
        self._doingLogin = False
        self._doingGetPerspective = False

    def clientConnectionFailed(self, connector, reason):
        super().clientConnectionFailed(connector, reason)
        # Twisted-1.3 erroneously abandons the connection on non-UserErrors.
        # To avoid this bug, don't upcall, and implement the correct version
        # of the method here.
        if self.continueTrying:
            self.connector = connector
            self.retry()

    def clientConnectionLost(self, connector, reason):
        super().clientConnectionLost(connector, reason, reconnecting=True)
        RCF = protocol.ReconnectingClientFactory
        RCF.clientConnectionLost(self, connector, reason)

    def clientConnectionMade(self, broker):
        self.resetDelay()
        super().clientConnectionMade(broker)
        if self._doingLogin:
            self.doLogin(self._root)
        if self._doingGetPerspective:
            self.doGetPerspective(self._root)
        self.gotRootObject(self._root)

    # oldcred methods

    def getPerspective(self, *args):
        raise RuntimeError("getPerspective is one-shot: use startGettingPerspective instead")

    def startGettingPerspective(
        self, username, password, serviceName, perspectiveName=None, client=None
    ):
        self._doingGetPerspective = True
        if perspectiveName is None:
            perspectiveName = username
        self._oldcredArgs = (username, password, serviceName, perspectiveName, client)

    def doGetPerspective(self, root):
        # oldcred getPerspective()
        (username, password, serviceName, perspectiveName, client) = self._oldcredArgs
        d = self._cbAuthIdentity(root, username, password)
        d.addCallback(self._cbGetPerspective, serviceName, perspectiveName, client)
        d.addCallbacks(self.gotPerspective, self.failedToGetPerspective)

    # newcred methods
    def login(self, *args):
        raise RuntimeError("login is one-shot: use startLogin instead")

    def startLogin(self, credentials, client=None):
        self._credentials = credentials
        self._client = client
        self._doingLogin = True

    def doLogin(self, root):
        # newcred login()
        d = self._cbSendUsername(
            root, self._credentials.username, self._credentials.password, self._client
        )
        d.addCallbacks(self.gotPerspective, self.failedToGetPerspective)

    # methods to override
    def gotPerspective(self, perspective):
        """The remote avatar or perspective (obtained each time this factory
        connects) is now available."""

    def gotRootObject(self, root):
        """The remote root object (obtained each time this factory connects)
        is now available. This method will be called each time the connection
        is established and the object reference is retrieved."""

    def failedToGetPerspective(self, why):
        """The login process failed, most likely because of an authorization
        failure (bad password), but it is also possible that we lost the new
        connection before we managed to send our credentials.
        """
        log.msg("ReconnectingPBClientFactory.failedToGetPerspective")
        if why.check(pb.PBConnectionLost):
            log.msg("we lost the brand-new connection")
            # retrying might help here, let clientConnectionLost decide
            return
        # probably authorization
        self.stopTrying()  # logging in harder won't help
        log.err(why)


def decode(data, encoding='utf-8', errors='strict'):
    """We need to convert a dictionary where keys and values
    are bytes, to unicode strings.  This happens when a
    Python 2 worker sends a dictionary back to a Python 3 master.
    """
    data_type = type(data)

    if data_type == bytes:
        return bytes2unicode(data, encoding, errors)
    if data_type in (dict, list, tuple):
        if data_type == dict:
            data = data.items()
        return data_type(map(decode, data))
    return data



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import error
from twisted.trial import unittest

from buildbot import config
from buildbot.process import remotetransfer
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import darcs
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectCpdir
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import sourcesteps


class TestDarcs(sourcesteps.SourceStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_no_empty_step_config(self):
        with self.assertRaises(config.ConfigErrors):
            darcs.Darcs()

    def test_incorrect_method(self):
        with self.assertRaises(config.ConfigErrors):
            darcs.Darcs(repourl='http://localhost/darcs', mode='full', method='fresh')

    def test_incremental_invalid_method(self):
        with self.assertRaises(config.ConfigErrors):
            darcs.Darcs(repourl='http://localhost/darcs', mode='incremental', method='fresh')

    def test_no_repo_url(self):
        with self.assertRaises(config.ConfigErrors):
            darcs.Darcs(mode='full', method='fresh')

    def test_mode_full_clobber(self):
        self.setup_step(
            darcs.Darcs(repourl='http://localhost/darcs', mode='full', method='clobber')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    'http://localhost/darcs',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_full_copy(self):
        self.setup_step(darcs.Darcs(repourl='http://localhost/darcs', mode='full', method='copy'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/_darcs', log_environ=True).exit(0),
            ExpectShell(workdir='source', command=['darcs', 'pull', '--all', '--verbose']).exit(0),
            ExpectCpdir(fromdir='source', todir='build', log_environ=True, timeout=1200).exit(0),
            ExpectShell(workdir='build', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_full_no_method(self):
        self.setup_step(darcs.Darcs(repourl='http://localhost/darcs', mode='full'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/_darcs', log_environ=True).exit(0),
            ExpectShell(workdir='source', command=['darcs', 'pull', '--all', '--verbose']).exit(0),
            ExpectCpdir(fromdir='source', todir='build', log_environ=True, timeout=1200).exit(0),
            ExpectShell(workdir='build', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_incremental(self):
        self.setup_step(darcs.Darcs(repourl='http://localhost/darcs', mode='incremental'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_darcs', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'pull', '--all', '--verbose']).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_incremental_patched(self):
        self.setup_step(darcs.Darcs(repourl='http://localhost/darcs', mode='incremental'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='source/_darcs', log_environ=True).exit(0),
            ExpectShell(workdir='source', command=['darcs', 'pull', '--all', '--verbose']).exit(0),
            ExpectCpdir(fromdir='source', todir='build', log_environ=True, timeout=1200).exit(0),
            ExpectStat(file='build/_darcs', log_environ=True).exit(0),
            ExpectShell(workdir='build', command=['darcs', 'pull', '--all', '--verbose']).exit(0),
            ExpectShell(workdir='build', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_incremental_patch(self):
        self.setup_step(
            darcs.Darcs(repourl='http://localhost/darcs', mode='incremental'), patch=(1, 'patch')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_darcs', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'pull', '--all', '--verbose']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_full_clobber_retry(self):
        self.setup_step(
            darcs.Darcs(
                repourl='http://localhost/darcs', mode='full', method='clobber', retry=(0, 2)
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    'http://localhost/darcs',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    'http://localhost/darcs',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    'http://localhost/darcs',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_full_clobber_revision(self):
        self.setup_step(
            darcs.Darcs(repourl='http://localhost/darcs', mode='full', method='clobber'),
            {"revision": 'abcdef01'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.darcs-context',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    '--context',
                    '.darcs-context',
                    'http://localhost/darcs',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_full_clobber_revision_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            darcs.Darcs(repourl='http://localhost/darcs', mode='full', method='clobber'),
            {"revision": 'abcdef01'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.darcs-context',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    '--context',
                    '.darcs-context',
                    'http://localhost/darcs',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_mode_incremental_no_existing_repo(self):
        self.setup_step(darcs.Darcs(repourl='http://localhost/darcs', mode='incremental'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectStat(file='wkdir/_darcs', log_environ=True).exit(1),
            ExpectShell(
                workdir='.',
                command=[
                    'darcs',
                    'get',
                    '--verbose',
                    '--lazy',
                    '--repo-name',
                    'wkdir',
                    'http://localhost/darcs',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['darcs', 'changes', '--max-count=1'])
            .stdout('Tue Aug 20 09:18:41 IST 2013 abc@gmail.com')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'Tue Aug 20 09:18:41 IST 2013 abc@gmail.com', 'Darcs')
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            darcs.Darcs(repourl='http://localhost/darcs', mode='full', method='clobber')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['darcs', '--version']).error(
                error.ConnectionLost()
            )
        )
        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Various decorators for test cases
"""

import os
import sys

from twisted.python import runtime

_FLAKY_ENV_VAR = 'RUN_FLAKY_TESTS'


def todo(message):
    """
    decorator to mark a todo test
    """

    def wrap(func):
        """
        just mark the test
        """
        func.todo = message
        return func

    return wrap


def flaky(bugNumber=None, issueNumber=None, onPlatform=None):
    def wrap(fn):
        if onPlatform is not None and sys.platform != onPlatform:
            return fn

        if os.environ.get(_FLAKY_ENV_VAR):
            return fn

        if bugNumber is not None:
            fn.skip = (
                f"Flaky test (http://trac.buildbot.net/ticket/{bugNumber}) "
                f"- set ${_FLAKY_ENV_VAR} to run anyway"
            )
        if issueNumber is not None:
            fn.skip = (
                f"Flaky test (https://github.com/buildbot/buildbot/issues/{issueNumber}) "
                f"- set ${_FLAKY_ENV_VAR} to run anyway"
            )
        return fn

    return wrap


def skipUnlessPlatformIs(platform):
    def closure(test):
        if runtime.platformType != platform:
            test.skip = f"not a {platform} platform"
        return test

    return closure


def skipIfPythonVersionIsLess(min_version_info):
    assert isinstance(min_version_info, tuple)

    def closure(test):
        if sys.version_info < min_version_info:
            test.skip = f"requires Python >= {min_version_info}"
        return test

    return closure



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake import fakemaster


class DBConnector_Basic(unittest.TestCase):
    """
    Basic tests of the DBConnector class - all start with an empty DB
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.master = yield fakemaster.make_master(
            self, wantRealReactor=True, wantDb=True, auto_upgrade=False, check_version=False
        )

    @defer.inlineCallbacks
    def test_is_current_empty(self):
        res = yield self.master.db.model.is_current()
        self.assertFalse(res)

    @defer.inlineCallbacks
    def test_is_current_full(self):
        yield self.master.db.model.upgrade()
        res = yield self.master.db.model.is_current()
        self.assertTrue(res)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass

import sqlalchemy as sa
from twisted.internet import defer
from twisted.python import deprecate
from twisted.python import versions

from buildbot.db import base
from buildbot.warnings import warn_deprecated


@dataclass
class TestResultModel:
    id: int
    builderid: int
    test_result_setid: int
    test_name: str | None
    test_code_path: str | None
    line: int | None
    duration_ns: int | None
    value: str | None

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'TestResultsConnectorComponent '
                'getTestResult, and getTestResults '
                'no longer return TestResult as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), TestResultModel)
class TestResultDict(TestResultModel):
    pass


class TestResultsConnectorComponent(base.DBConnectorComponent):
    def _add_code_paths(self, builderid: int, paths: set[str]) -> defer.Deferred[dict[str, int]]:
        # returns a dictionary of path to id in the test_code_paths table.
        # For paths that already exist, the id of the row in the test_code_paths is retrieved.
        assert isinstance(paths, set)

        def thd(conn) -> dict[str, int]:
            paths_to_ids = {}
            paths_table = self.db.model.test_code_paths

            for path_batch in self.doBatch(paths, batch_n=3000):
                path_batch = set(path_batch)

                while path_batch:
                    # Use expanding bindparam, because performance of sqlalchemy is very slow
                    # when filtering large sets otherwise.
                    q = paths_table.select().where(
                        (paths_table.c.path.in_(sa.bindparam('paths', expanding=True)))
                        & (paths_table.c.builderid == builderid)
                    )

                    res = conn.execute(q, {'paths': list(path_batch)})
                    for row in res.fetchall():
                        paths_to_ids[row.path] = row.id
                        path_batch.remove(row.path)

                    # paths now contains all the paths that need insertion.
                    try:
                        insert_values = [
                            {'builderid': builderid, 'path': path} for path in path_batch
                        ]

                        q = paths_table.insert().values(insert_values)

                        if self.db.pool.engine.dialect.name in ['postgresql', 'mssql']:
                            # Use RETURNING, this way we won't need an additional select query
                            q = q.returning(paths_table.c.id, paths_table.c.path)

                            res = conn.execute(q)
                            conn.commit()
                            for row in res.fetchall():
                                paths_to_ids[row.path] = row.id
                                path_batch.remove(row.path)
                        else:
                            conn.execute(q)
                            conn.commit()

                    except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                        # There was a competing addCodePaths() call that added a path for the same
                        # builder. Depending on the DB driver, none or some rows were inserted, but
                        # we will re-check what's got inserted in the next iteration of the loop
                        conn.rollback()

            return paths_to_ids

        return self.db.pool.do(thd)

    def getTestCodePaths(
        self, builderid, path_prefix: str | None = None, result_spec=None
    ) -> defer.Deferred[list[str]]:
        def thd(conn) -> list[str]:
            paths_table = self.db.model.test_code_paths
            q = paths_table.select()
            if path_prefix is not None:
                q = q.where(paths_table.c.path.startswith(path_prefix))
            if result_spec is not None:
                return result_spec.thd_execute(conn, q, lambda x: x['path'])
            res = conn.execute(q)
            return [row.path for row in res.fetchall()]

        return self.db.pool.do(thd)

    def _add_names(self, builderid: int, names: set[str]) -> defer.Deferred[dict[str, int]]:
        # returns a dictionary of name to id in the test_names table.
        # For names that already exist, the id of the row in the test_names is retrieved.
        assert isinstance(names, set)

        def thd(conn) -> dict[str, int]:
            names_to_ids = {}
            names_table = self.db.model.test_names

            for name_batch in self.doBatch(names, batch_n=3000):
                name_batch = set(name_batch)
                while name_batch:
                    # Use expanding bindparam, because performance of sqlalchemy is very slow
                    # when filtering large sets otherwise.
                    q = names_table.select().where(
                        (names_table.c.name.in_(sa.bindparam('names', expanding=True)))
                        & (names_table.c.builderid == builderid)
                    )

                    res = conn.execute(q, {'names': list(name_batch)})
                    for row in res.fetchall():
                        names_to_ids[row.name] = row.id
                        name_batch.remove(row.name)

                    # names now contains all the names that need insertion.
                    try:
                        insert_values = [
                            {'builderid': builderid, 'name': name} for name in name_batch
                        ]

                        q = names_table.insert().values(insert_values)

                        if self.db.pool.engine.dialect.name in ['postgresql', 'mssql']:
                            # Use RETURNING, this way we won't need an additional select query
                            q = q.returning(names_table.c.id, names_table.c.name)

                            res = conn.execute(q)
                            conn.commit()
                            for row in res.fetchall():
                                names_to_ids[row.name] = row.id
                                name_batch.remove(row.name)
                        else:
                            conn.execute(q)
                            conn.commit()

                    except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                        # There was a competing addNames() call that added a name for the same
                        # builder. Depending on the DB driver, none or some rows were inserted, but
                        # we will re-check what's got inserted in the next iteration of the loop
                        conn.rollback()

            return names_to_ids

        return self.db.pool.do(thd)

    def getTestNames(
        self, builderid, name_prefix=None, result_spec=None
    ) -> defer.Deferred[list[str]]:
        def thd(conn) -> list[str]:
            names_table = self.db.model.test_names
            q = names_table.select().where(names_table.c.builderid == builderid)
            if name_prefix is not None:
                q = q.where(names_table.c.name.startswith(name_prefix))
            if result_spec is not None:
                return result_spec.thd_execute(conn, q, lambda x: x.name)
            res = conn.execute(q)
            return [row.name for row in res.fetchall()]

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def addTestResults(self, builderid, test_result_setid, result_values):
        # Adds multiple test results for a specific test result set.
        # result_values is a list of dictionaries each of which must contain 'value' key and at
        # least one of 'test_name', 'test_code_path'. 'line' key is optional.
        # The function returns nothing.

        # Build values list for insertion.
        insert_values = []

        insert_names = set()
        insert_code_paths = set()
        for result_value in result_values:
            if 'value' not in result_value:
                raise KeyError('Each of result_values must contain \'value\' key')

            if 'test_name' not in result_value and 'test_code_path' not in result_value:
                raise KeyError(
                    'Each of result_values must contain at least one of '
                    '\'test_name\' or \'test_code_path\' keys'
                )

            if 'test_name' in result_value:
                insert_names.add(result_value['test_name'])
            if 'test_code_path' in result_value:
                insert_code_paths.add(result_value['test_code_path'])

        code_path_to_id = yield self._add_code_paths(builderid, insert_code_paths)
        name_to_id = yield self._add_names(builderid, insert_names)

        for result_value in result_values:
            insert_value = {
                'value': result_value['value'],
                'builderid': builderid,
                'test_result_setid': test_result_setid,
                'test_nameid': None,
                'test_code_pathid': None,
                'line': None,
                'duration_ns': None,
            }

            if 'test_name' in result_value:
                insert_value['test_nameid'] = name_to_id[result_value['test_name']]
            if 'test_code_path' in result_value:
                insert_value['test_code_pathid'] = code_path_to_id[result_value['test_code_path']]
            if 'line' in result_value:
                insert_value['line'] = result_value['line']
            if 'duration_ns' in result_value:
                insert_value['duration_ns'] = result_value['duration_ns']

            insert_values.append(insert_value)

        def thd(conn):
            results_table = self.db.model.test_results
            q = results_table.insert().values(insert_values)
            conn.execute(q)

        yield self.db.pool.do_with_transaction(thd)

    def getTestResult(self, test_resultid: int) -> defer.Deferred[TestResultModel | None]:
        def thd(conn) -> TestResultModel | None:
            results_table = self.db.model.test_results
            code_paths_table = self.db.model.test_code_paths
            names_table = self.db.model.test_names

            j = results_table.outerjoin(code_paths_table).outerjoin(names_table)

            q = sa.select(results_table, code_paths_table.c.path, names_table.c.name)
            q = q.select_from(j).where(results_table.c.id == test_resultid)

            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            return self._mode_from_row(row)

        return self.db.pool.do(thd)

    def getTestResults(
        self, builderid: int, test_result_setid: int, result_spec=None
    ) -> defer.Deferred[list[TestResultModel]]:
        def thd(conn) -> list[TestResultModel]:
            results_table = self.db.model.test_results
            code_paths_table = self.db.model.test_code_paths
            names_table = self.db.model.test_names

            # specify join ON clauses manually to force filtering of code_paths_table and
            # names_table before join

            j = results_table.outerjoin(
                code_paths_table,
                (results_table.c.test_code_pathid == code_paths_table.c.id)
                & (code_paths_table.c.builderid == builderid),
            )

            j = j.outerjoin(
                names_table,
                (results_table.c.test_nameid == names_table.c.id)
                & (names_table.c.builderid == builderid),
            )

            q = sa.select(results_table, code_paths_table.c.path, names_table.c.name)
            q = q.select_from(j).where(
                (results_table.c.builderid == builderid)
                & (results_table.c.test_result_setid == test_result_setid)
            )

            if result_spec is not None:
                return result_spec.thd_execute(conn, q, self._mode_from_row)
            res = conn.execute(q)
            return [self._mode_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    def _mode_from_row(self, row):
        return TestResultModel(
            id=row.id,
            builderid=row.builderid,
            test_result_setid=row.test_result_setid,
            test_name=row.name,
            test_code_path=row.path,
            line=row.line,
            duration_ns=row.duration_ns,
            value=row.value,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.trial import unittest

from buildbot_worker.commands import shell
from buildbot_worker.test.fake.runprocess import Expect
from buildbot_worker.test.util.command import CommandTestMixin

if TYPE_CHECKING:
    from buildbot_worker.util.twisted import InlineCallbacksType


class TestWorkerShellCommand(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_simple(self) -> InlineCallbacksType[None]:
        workdir = os.path.join(self.basedir, 'workdir')
        self.make_command(
            shell.WorkerShellCommand, {'command': ['echo', 'hello'], 'workdir': workdir}
        )

        self.patch_runprocess(
            Expect(['echo', 'hello'], self.basedir_workdir)
            .update('header', 'headers')
            .update('stdout', 'hello\n')
            .update('rc', 0)
            .exit(0)
        )

        yield self.run_command()

        # note that WorkerShellCommand does not add any extra updates of it own
        self.assertUpdates(
            [('header', 'headers'), ('stdout', 'hello\n'), ('rc', 0)], self.protocol_command.show()
        )

    # TODO: test all functionality that WorkerShellCommand adds atop RunProcess



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
# Copyright Mamba Team

from io import BytesIO

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake.web import FakeRequest
from buildbot.test.fake.web import fakeMasterForHooks
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import unicode2bytes
from buildbot.www import change_hook
from buildbot.www.hooks.bitbucketserver import _HEADER_EVENT

_CT_JSON = b'application/json'

bitbucketPRproperties = {
    'pullrequesturl': 'http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21',
    'bitbucket.id': '21',
    'bitbucket.link': 'http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21',
    'bitbucket.title': 'dot 1496311906',
    'bitbucket.authorLogin': 'Buildbot',
    'bitbucket.fromRef.branch.name': 'branch_1496411680',
    'bitbucket.fromRef.branch.rawNode': 'a87e21f7433d8c16ac7be7413483fbb76c72a8ba',
    'bitbucket.fromRef.commit.authorTimestamp': 0,
    'bitbucket.fromRef.commit.date': None,
    'bitbucket.fromRef.commit.hash': 'a87e21f7433d8c16ac7be7413483fbb76c72a8ba',
    'bitbucket.fromRef.commit.message': None,
    'bitbucket.fromRef.repository.fullName': 'CI/py-repo',
    'bitbucket.fromRef.repository.links.self': [
        {'href': 'http://localhost:7990/projects/CI/repos/py-repo/browse'}
    ],
    'bitbucket.fromRef.repository.owner.displayName': 'CI',
    'bitbucket.fromRef.repository.owner.username': 'CI',
    'bitbucket.fromRef.repository.ownerName': 'CI',
    'bitbucket.fromRef.repository.project.key': 'CI',
    'bitbucket.fromRef.repository.project.name': 'Continuous Integration',
    'bitbucket.fromRef.repository.public': False,
    'bitbucket.fromRef.repository.scmId': 'git',
    'bitbucket.fromRef.repository.slug': 'py-repo',
    'bitbucket.toRef.branch.name': 'master',
    'bitbucket.toRef.branch.rawNode': '7aebbb0089c40fce138a6d0b36d2281ea34f37f5',
    'bitbucket.toRef.commit.authorTimestamp': 0,
    'bitbucket.toRef.commit.date': None,
    'bitbucket.toRef.commit.hash': '7aebbb0089c40fce138a6d0b36d2281ea34f37f5',
    'bitbucket.toRef.commit.message': None,
    'bitbucket.toRef.repository.fullName': 'CI/py-repo',
    'bitbucket.toRef.repository.links.self': [
        {'href': 'http://localhost:7990/projects/CI/repos/py-repo/browse'}
    ],
    'bitbucket.toRef.repository.owner.displayName': 'CI',
    'bitbucket.toRef.repository.owner.username': 'CI',
    'bitbucket.toRef.repository.ownerName': 'CI',
    'bitbucket.toRef.repository.project.key': 'CI',
    'bitbucket.toRef.repository.project.name': 'Continuous Integration',
    'bitbucket.toRef.repository.public': False,
    'bitbucket.toRef.repository.scmId': 'git',
    'bitbucket.toRef.repository.slug': 'py-repo',
}


pushJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "public": false,
        "ownerName": "CI",
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    },
    "push": {
        "changes": [
            {
                "created": false,
                "closed": false,
                "new": {
                    "type": "branch",
                    "name": "branch_1496411680",
                    "target": {
                        "type": "commit",
                        "hash": "793d4754230023d85532f9a38dba3290f959beb4"
                    }
                },
                "old": {
                    "type": "branch",
                    "name": "branch_1496411680",
                    "target": {
                        "type": "commit",
                        "hash": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba"
                    }
                }
            }
        ]
    }
}
"""

pullRequestCreatedJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "pullrequest": {
        "id": "21",
        "title": "dot 1496311906",
        "link": "http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21",
        "authorLogin": "Buildbot",
        "fromRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "name": "branch_1496411680"
            }
        },
        "toRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "name": "master"
            }
        }
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "public": false,
        "ownerName": "CI",
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    }
}
"""

pullRequestUpdatedJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "pullrequest": {
        "id": "21",
        "title": "dot 1496311906",
        "link": "http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21",
        "authorLogin": "Buildbot",
        "fromRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "name": "branch_1496411680"
            }
        },
        "toRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "name": "master"
            }
        }
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "public": false,
        "ownerName": "CI",
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    }
}
"""

pullRequestRejectedJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "pullrequest": {
        "id": "21",
        "title": "dot 1496311906",
        "link": "http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21",
        "authorLogin": "Buildbot",
        "fromRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "name": "branch_1496411680"
            }
        },
        "toRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "name": "master"
            }
        }
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "public": false,
        "ownerName": "CI",
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    }
}
"""

pullRequestFulfilledJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "pullrequest": {
        "id": "21",
        "title": "dot 1496311906",
        "link": "http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21",
        "authorLogin": "Buildbot",
        "fromRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "a87e21f7433d8c16ac7be7413483fbb76c72a8ba",
                "name": "branch_1496411680"
            }
        },
        "toRef": {
            "repository": {
                "scmId": "git",
                "project": {
                    "key": "CI",
                    "name": "Continuous Integration"
                },
                "slug": "py-repo",
                "links": {
                    "self": [
                        {
                            "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                        }
                    ]
                },
                "public": false,
                "ownerName": "CI",
                "owner": {
                    "username": "CI",
                    "displayName": "CI"
                },
                "fullName": "CI/py-repo"
            },
            "commit": {
                "message": null,
                "date": null,
                "hash": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "authorTimestamp": 0
            },
            "branch": {
                "rawNode": "7aebbb0089c40fce138a6d0b36d2281ea34f37f5",
                "name": "master"
            }
        }
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "public": false,
        "ownerName": "CI",
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    }
}
"""

deleteTagJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "ownerName": "BUIL",
        "public": false,
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    },
    "push": {
        "changes": [
            {
                "created": false,
                "closed": true,
                "old": {
                    "type": "tag",
                    "name": "1.0.0",
                    "target": {
                        "type": "commit",
                        "hash": "793d4754230023d85532f9a38dba3290f959beb4"
                    }
                },
                "new": null
            }
        ]
    }
}
"""

deleteBranchJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "ownerName": "CI",
        "public": false,
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    },
    "push": {
        "changes": [
            {
                "created": false,
                "closed": true,
                "old": {
                    "type": "branch",
                    "name": "branch_1496758965",
                    "target": {
                        "type": "commit",
                        "hash": "793d4754230023d85532f9a38dba3290f959beb4"
                    }
                },
                "new": null
            }
        ]
    }
}
"""

newTagJsonPayload = """
{
    "actor": {
        "username": "John",
        "displayName": "John Smith"
    },
    "repository": {
        "scmId": "git",
        "project": {
            "key": "CI",
            "name": "Continuous Integration"
        },
        "slug": "py-repo",
        "links": {
            "self": [
                {
                    "href": "http://localhost:7990/projects/CI/repos/py-repo/browse"
                }
            ]
        },
        "public": false,
        "ownerName": "CI",
        "owner": {
            "username": "CI",
            "displayName": "CI"
        },
        "fullName": "CI/py-repo"
    },
    "push": {
        "changes": [
            {
                "created": true,
                "closed": false,
                "old": null,
                "new": {
                    "type": "tag",
                    "name": "1.0.0",
                    "target": {
                        "type": "commit",
                        "hash": "793d4754230023d85532f9a38dba3290f959beb4"
                    }
                }
            }
        ]
    }
}
"""


def _prepare_request(payload, headers=None, change_dict=None):
    headers = headers or {}
    request = FakeRequest(change_dict)
    request.uri = b"/change_hook/bitbucketserver"
    request.method = b"POST"
    if isinstance(payload, str):
        payload = unicode2bytes(payload)
    request.content = BytesIO(payload)
    request.received_headers[b'Content-Type'] = _CT_JSON
    request.received_headers.update(headers)
    return request


class TestChangeHookConfiguredWithGitChange(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        master = yield fakeMasterForHooks(self)
        self.change_hook = change_hook.ChangeHookResource(
            dialects={
                'bitbucketserver': {
                    'bitbucket_property_whitelist': ["bitbucket.*"],
                }
            },
            master=master,
        )

    def assertDictSubset(self, expected_dict, response_dict):
        expected = {}
        for key in expected_dict.keys():
            self.assertIn(key, set(response_dict.keys()))
            expected[key] = response_dict[key]
        self.assertDictEqual(expected_dict, expected)

    def _checkPush(self, change):
        self.assertEqual(change['repository'], 'http://localhost:7990/projects/CI/repos/py-repo/')
        self.assertEqual(change['author'], 'John Smith <John>')
        self.assertEqual(change['project'], 'Continuous Integration')
        self.assertEqual(change['revision'], '793d4754230023d85532f9a38dba3290f959beb4')
        self.assertEqual(
            change['comments'],
            'Bitbucket Server commit 793d4754230023d85532f9a38dba3290f959beb4',
        )
        self.assertEqual(
            change['revlink'],
            'http://localhost:7990/projects/CI/repos/py-repo/commits/'
            '793d4754230023d85532f9a38dba3290f959beb4',
        )

    @defer.inlineCallbacks
    def testHookWithChangeOnRefsChangedEvent(self):
        request = _prepare_request(pushJsonPayload, headers={_HEADER_EVENT: 'repo:refs_changed'})

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPush(change)
        self.assertEqual(change['branch'], 'refs/heads/branch_1496411680')
        self.assertEqual(change['category'], 'push')

    @defer.inlineCallbacks
    def testHookWithChangeOnPushEvent(self):
        request = _prepare_request(pushJsonPayload, headers={_HEADER_EVENT: 'repo:push'})

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPush(change)
        self.assertEqual(change['branch'], 'refs/heads/branch_1496411680')
        self.assertEqual(change['category'], 'push')

    @defer.inlineCallbacks
    def testHookWithNonDictOption(self):
        self.change_hook.dialects = {'bitbucketserver': True}
        yield self.testHookWithChangeOnPushEvent()

    def _checkPullRequest(self, change):
        self.assertEqual(change['repository'], 'http://localhost:7990/projects/CI/repos/py-repo/')
        self.assertEqual(change['author'], 'John Smith <John>')
        self.assertEqual(change['project'], 'Continuous Integration')
        self.assertEqual(change['comments'], 'Bitbucket Server Pull Request #21')
        self.assertEqual(
            change['revlink'], 'http://localhost:7990/projects/CI/repos/py-repo/pull-requests/21'
        )
        self.assertEqual(change['revision'], 'a87e21f7433d8c16ac7be7413483fbb76c72a8ba')
        self.assertDictSubset(bitbucketPRproperties, change["properties"])

    @defer.inlineCallbacks
    def testHookWithChangeOnPullRequestCreated(self):
        request = _prepare_request(
            pullRequestCreatedJsonPayload, headers={_HEADER_EVENT: 'pullrequest:created'}
        )

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPullRequest(change)
        self.assertEqual(change['branch'], 'refs/pull-requests/21/merge')
        self.assertEqual(change['category'], 'pull-created')

    @defer.inlineCallbacks
    def testHookWithChangeOnPullRequestUpdated(self):
        request = _prepare_request(
            pullRequestUpdatedJsonPayload, headers={_HEADER_EVENT: 'pullrequest:updated'}
        )

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPullRequest(change)
        self.assertEqual(change['branch'], 'refs/pull-requests/21/merge')
        self.assertEqual(change['category'], 'pull-updated')

    @defer.inlineCallbacks
    def testHookWithChangeOnPullRequestRejected(self):
        request = _prepare_request(
            pullRequestRejectedJsonPayload, headers={_HEADER_EVENT: 'pullrequest:rejected'}
        )

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPullRequest(change)
        self.assertEqual(change['branch'], 'refs/heads/branch_1496411680')
        self.assertEqual(change['category'], 'pull-rejected')

    @defer.inlineCallbacks
    def testHookWithChangeOnPullRequestFulfilled(self):
        request = _prepare_request(
            pullRequestFulfilledJsonPayload, headers={_HEADER_EVENT: 'pullrequest:fulfilled'}
        )

        yield request.test_render(self.change_hook)

        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPullRequest(change)
        self.assertEqual(change['branch'], 'refs/heads/master')
        self.assertEqual(change['category'], 'pull-fulfilled')

    @defer.inlineCallbacks
    def _checkCodebase(self, event_type, expected_codebase):
        payloads = {
            'repo:refs_changed': pushJsonPayload,
            'pullrequest:updated': pullRequestUpdatedJsonPayload,
        }
        request = _prepare_request(payloads[event_type], headers={_HEADER_EVENT: event_type})
        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self.assertEqual(change['codebase'], expected_codebase)

    @defer.inlineCallbacks
    def testHookWithCodebaseValueOnPushEvent(self):
        self.change_hook.dialects = {'bitbucketserver': {'codebase': 'super-codebase'}}
        yield self._checkCodebase('repo:refs_changed', 'super-codebase')

    @defer.inlineCallbacks
    def testHookWithCodebaseFunctionOnPushEvent(self):
        self.change_hook.dialects = {
            'bitbucketserver': {'codebase': lambda payload: payload['repository']['project']['key']}
        }
        yield self._checkCodebase('repo:refs_changed', 'CI')

    @defer.inlineCallbacks
    def testHookWithCodebaseValueOnPullEvent(self):
        self.change_hook.dialects = {'bitbucketserver': {'codebase': 'super-codebase'}}
        yield self._checkCodebase('pullrequest:updated', 'super-codebase')

    @defer.inlineCallbacks
    def testHookWithCodebaseFunctionOnPullEvent(self):
        self.change_hook.dialects = {
            'bitbucketserver': {'codebase': lambda payload: payload['repository']['project']['key']}
        }
        yield self._checkCodebase('pullrequest:updated', 'CI')

    @defer.inlineCallbacks
    def testHookWithUnhandledEvent(self):
        request = _prepare_request(pushJsonPayload, headers={_HEADER_EVENT: 'invented:event'})
        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 0)
        self.assertEqual(request.written, b"Unknown event: invented_event")

    @defer.inlineCallbacks
    def testHookWithChangeOnCreateTag(self):
        request = _prepare_request(newTagJsonPayload, headers={_HEADER_EVENT: 'repo:refs_changed'})
        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPush(change)
        self.assertEqual(change['branch'], 'refs/tags/1.0.0')
        self.assertEqual(change['category'], 'push')

    @defer.inlineCallbacks
    def testHookWithChangeOnDeleteTag(self):
        request = _prepare_request(
            deleteTagJsonPayload, headers={_HEADER_EVENT: 'repo:refs_changed'}
        )
        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPush(change)
        self.assertEqual(change['branch'], 'refs/tags/1.0.0')
        self.assertEqual(change['category'], 'ref-deleted')

    @defer.inlineCallbacks
    def testHookWithChangeOnDeleteBranch(self):
        request = _prepare_request(
            deleteBranchJsonPayload, headers={_HEADER_EVENT: 'repo:refs_changed'}
        )
        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 1)
        change = self.change_hook.master.data.updates.changesAdded[0]
        self._checkPush(change)
        self.assertEqual(change['branch'], 'refs/heads/branch_1496758965')
        self.assertEqual(change['category'], 'ref-deleted')

    @defer.inlineCallbacks
    def testHookWithInvalidContentType(self):
        request = _prepare_request(pushJsonPayload, headers={_HEADER_EVENT: b'repo:refs_changed'})
        request.received_headers[b'Content-Type'] = b'invalid/content'
        yield request.test_render(self.change_hook)
        self.assertEqual(len(self.change_hook.master.data.updates.changesAdded), 0)
        self.assertEqual(request.written, b"Unknown content type: 'invalid/content'")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

if TYPE_CHECKING:
    from buildbot.db.steps import StepModel
    from buildbot.util.twisted import InlineCallbacksType


def _db2data(model: StepModel):
    return {
        'stepid': model.id,
        'number': model.number,
        'name': model.name,
        'buildid': model.buildid,
        'started_at': model.started_at,
        "locks_acquired_at": model.locks_acquired_at,
        'complete': model.complete_at is not None,
        'complete_at': model.complete_at,
        'state_string': model.state_string,
        'results': model.results,
        'urls': [{'name': item.name, 'url': item.url} for item in model.urls],
        'hidden': model.hidden,
    }


class StepEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/steps/n:stepid",
        "/builds/n:buildid/steps/i:step_name",
        "/builds/n:buildid/steps/n:step_number",
        "/builders/n:builderid/builds/n:build_number/steps/i:step_name",
        "/builders/n:builderid/builds/n:build_number/steps/n:step_number",
        "/builders/s:buildername/builds/n:build_number/steps/i:step_name",
        "/builders/s:buildername/builds/n:build_number/steps/n:step_number",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        if 'stepid' in kwargs:
            dbdict = yield self.master.db.steps.getStep(kwargs['stepid'])
            return _db2data(dbdict) if dbdict else None
        buildid = yield self.getBuildid(kwargs)
        if buildid is None:
            return None
        dbdict = yield self.master.db.steps.getStep(
            buildid=buildid, number=kwargs.get('step_number'), name=kwargs.get('step_name')
        )
        return _db2data(dbdict) if dbdict else None


class StepsEndpoint(base.BuildNestingMixin, base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/builds/n:buildid/steps",
        "/builders/n:builderid/builds/n:build_number/steps",
        "/builders/s:buildername/builds/n:build_number/steps",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        if 'buildid' in kwargs:
            buildid = kwargs['buildid']
        else:
            buildid = yield self.getBuildid(kwargs)
            if buildid is None:
                return None
        steps = yield self.master.db.steps.getSteps(buildid=buildid)
        return [_db2data(model) for model in steps]


class UrlEntityType(types.Entity):
    name = types.String()
    url = types.String()


class Step(base.ResourceType):
    name = "step"
    plural = "steps"
    endpoints = [StepEndpoint, StepsEndpoint]
    eventPathPatterns = [
        "/builds/:buildid/steps/:stepid",
        "/steps/:stepid",
    ]

    class EntityType(types.Entity):
        stepid = types.Integer()
        number = types.Integer()
        name = types.Identifier(50)
        buildid = types.Integer()
        started_at = types.NoneOk(types.DateTime())
        locks_acquired_at = types.NoneOk(types.DateTime())
        complete = types.Boolean()
        complete_at = types.NoneOk(types.DateTime())
        results = types.NoneOk(types.Integer())
        state_string = types.String()
        urls = types.List(of=UrlEntityType("Url"))
        hidden = types.Boolean()

    entityType = EntityType(name)

    @defer.inlineCallbacks
    def generateEvent(self, stepid, event):
        step = yield self.master.data.get(('steps', stepid))
        self.produceEvent(step, event)

    @base.updateMethod
    @defer.inlineCallbacks
    def addStep(self, buildid, name):
        stepid, num, name = yield self.master.db.steps.addStep(
            buildid=buildid, name=name, state_string='pending'
        )
        yield self.generateEvent(stepid, 'new')
        return (stepid, num, name)

    @base.updateMethod
    @defer.inlineCallbacks
    def startStep(self, stepid: int, started_at: int | None = None, locks_acquired: bool = False):
        if started_at is None:
            started_at = int(self.master.reactor.seconds())
        yield self.master.db.steps.startStep(
            stepid=stepid, started_at=started_at, locks_acquired=locks_acquired
        )
        yield self.generateEvent(stepid, 'started')

    @base.updateMethod
    @defer.inlineCallbacks
    def set_step_locks_acquired_at(self, stepid: int, locks_acquired_at: int | None = None):
        if locks_acquired_at is None:
            locks_acquired_at = int(self.master.reactor.seconds())

        yield self.master.db.steps.set_step_locks_acquired_at(
            stepid=stepid, locks_acquired_at=locks_acquired_at
        )
        yield self.generateEvent(stepid, 'updated')

    @base.updateMethod
    @defer.inlineCallbacks
    def setStepStateString(self, stepid: int, state_string: str) -> InlineCallbacksType[None]:
        yield self.master.db.steps.setStepStateString(stepid=stepid, state_string=state_string)
        yield self.generateEvent(stepid, 'updated')

    @base.updateMethod
    @defer.inlineCallbacks
    def addStepURL(self, stepid: int, name: str, url: str) -> InlineCallbacksType[None]:
        yield self.master.db.steps.addURL(stepid=stepid, name=name, url=url)
        yield self.generateEvent(stepid, 'updated')

    @base.updateMethod
    @defer.inlineCallbacks
    def finishStep(self, stepid: int, results: int, hidden: bool) -> InlineCallbacksType[None]:
        yield self.master.db.steps.finishStep(stepid=stepid, results=results, hidden=hidden)
        yield self.generateEvent(stepid, 'finished')



import json

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake.web import FakeRequest
from buildbot.test.fake.web import fakeMasterForHooks
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import bytes2unicode
from buildbot.www.change_hook import ChangeHookResource
from buildbot.www.hooks.base import BaseHookHandler


@defer.inlineCallbacks
def _prepare_base_change_hook(testcase, **options):
    master = yield fakeMasterForHooks(testcase)
    return ChangeHookResource(dialects={'base': options}, master=master)


def _prepare_request(payload, headers=None):
    if headers is None:
        headers = {b"Content-type": b"application/x-www-form-urlencoded", b"Accept": b"text/plain"}
    else:
        headers = {}

    if b'comments' not in payload:
        payload[b'comments'] = b'test_www_hook_base submission'  # Required field

    request = FakeRequest()

    request.uri = b"/change_hook/base"
    request.method = b"POST"
    request.args = payload
    request.received_headers.update(headers)

    return request


class TestChangeHookConfiguredWithBase(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.changeHook = yield _prepare_base_change_hook(self)

    @defer.inlineCallbacks
    def _check_base_with_change(self, payload):
        self.request = _prepare_request(payload)
        yield self.request.test_render(self.changeHook)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 1)
        change = self.changeHook.master.data.updates.changesAdded[0]

        def _first_or_nothing(val):
            if isinstance(val, type([])):
                val = val[0]
            return bytes2unicode(val)

        if payload.get(b'files'):
            files = json.loads(_first_or_nothing(payload.get(b'files')))
        else:
            files = []
        self.assertEqual(change['files'], files)

        if payload.get(b'properties'):
            props = json.loads(_first_or_nothing(payload.get(b'properties')))
        else:
            props = {}
        self.assertEqual(change['properties'], props)

        self.assertEqual(
            change['author'], _first_or_nothing(payload.get(b'author', payload.get(b'who')))
        )

        for field in ('revision', 'committer', 'comments', 'branch', 'category', 'revlink'):
            self.assertEqual(change[field], _first_or_nothing(payload.get(field.encode())))

        for field in ('repository', 'project'):
            self.assertEqual(change[field], _first_or_nothing(payload.get(field.encode())) or '')

    def test_base_with_no_change(self):
        return self._check_base_with_change({})

    def test_base_with_changes(self):
        return self._check_base_with_change({
            b'revision': [b'1234badcaca5678'],
            b'branch': [b'master'],
            b'comments': [b'Fix foo bar'],
            b'category': [b'bug'],
            b'revlink': [b'https://git.myproject.org/commit/1234badcaca5678'],
            b'repository': [b'myproject'],
            b'project': [b'myproject'],
            b'author': [b'me <me@myself.org>'],
            b'committer': [b'me <me@myself.org>'],
            b'files': [b'["src/main.c", "src/foo.c"]'],
            b'properties': [b'{"color": "blue", "important": true, "size": 2}'],
        })


class TestChangeHookConfiguredWithCustomBase(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()

        class CustomBase(BaseHookHandler):
            def getChanges(self, request):
                args = request.args
                chdict = {
                    "revision": args.get(b'revision'),
                    "repository": args.get(b'_repository') or '',
                    "project": args.get(b'project') or '',
                    "codebase": args.get(b'codebase'),
                }
                return ([chdict], None)

        self.changeHook = yield _prepare_base_change_hook(self, custom_class=CustomBase)

    @defer.inlineCallbacks
    def _check_base_with_change(self, payload):
        self.request = _prepare_request(payload)
        yield self.request.test_render(self.changeHook)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 1)
        change = self.changeHook.master.data.updates.changesAdded[0]
        self.assertEqual(change['repository'], payload.get(b'_repository') or '')

    def test_base_with_no_change(self):
        return self._check_base_with_change({b'repository': b'foo'})



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.trial import unittest

from buildbot.test.util import migration
from buildbot.util import sautils

if TYPE_CHECKING:
    from twisted.internet import defer


class Migration(migration.MigrateTestMixin, unittest.TestCase):
    def setUp(self) -> defer.Deferred[None]:  # type: ignore[override]
        return self.setUpMigrateTest()

    def create_tables_thd(self, conn: sa.future.engine.Connection) -> None:
        metadata = sa.MetaData()
        metadata.bind = conn  # type: ignore[attr-defined]

        projects_tbl = sautils.Table(
            'projects',
            metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('name', sa.Text, nullable=False),
            sa.Column('name_hash', sa.String(40), nullable=False),
            sa.Column('slug', sa.String(50), nullable=False),
            sa.Column('description', sa.Text, nullable=True),
            sa.Column('description_format', sa.Text, nullable=True),
            sa.Column('description_html', sa.Text, nullable=True),
        )
        projects_tbl.create(bind=conn)

        conn.commit()

    def test_update(self) -> defer.Deferred[None]:
        def setup_thd(conn: sa.future.engine.Connection) -> None:
            self.create_tables_thd(conn)

        def verify_thd(conn: sa.future.engine.Connection) -> None:
            metadata = sa.MetaData()
            metadata.bind = conn  # type: ignore[attr-defined]

            # verify tables have been created
            codebases = sautils.Table('codebases', metadata, autoload_with=conn)
            q = sa.select(
                codebases.c.projectid,
                codebases.c.name,
                codebases.c.name_hash,
                codebases.c.slug,
            )
            self.assertEqual(conn.execute(q).fetchall(), [])

            codebase_commits = sautils.Table('codebase_commits', metadata, autoload_with=conn)
            q = sa.select(
                codebase_commits.c.codebaseid,
                codebase_commits.c.author,
                codebase_commits.c.committer,
                codebase_commits.c.comments,
                codebase_commits.c.when_timestamp,
                codebase_commits.c.revision,
                codebase_commits.c.parent_commitid,
            )
            self.assertEqual(conn.execute(q).fetchall(), [])

            codebase_branches = sautils.Table('codebase_branches', metadata, autoload_with=conn)
            q = sa.select(
                codebase_branches.c.codebaseid,
                codebase_branches.c.name,
                codebase_branches.c.name_hash,
                codebase_branches.c.commitid,
                codebase_branches.c.last_timestamp,
            )
            self.assertEqual(conn.execute(q).fetchall(), [])

            # verify indexes have been created
            insp = sa.inspect(conn)

            def verify_indexes(table: str, expected_indexes: list[str]) -> None:
                indexes = insp.get_indexes(table)
                index_names = [item['name'] for item in indexes]
                for expected_index in expected_indexes:
                    self.assertTrue(expected_index in index_names)

            verify_indexes('codebases', ['codebases_projects_name_hash'])
            verify_indexes('codebase_branches', ['codebase_branches_unique'])
            verify_indexes('codebase_commits', ['codebase_commits_unique'])

        return self.do_test_migration('066', '067', setup_thd, verify_thd)
